@@ -1,14 +1,36 @@
 _NOW = 1151365354
 _TIMEFORMAT = '%b %d %I:%M %p'
 
-class DummyOptions:
+import functools
 
-    make_pipes_error = None
-    fork_error = None
-    execv_error = None
-    kill_error = None
-    minfds = 5
+from supervisor.compat import Fault
+from supervisor.compat import as_bytes
+
+# mock is imported here for py2/3 compat.  we only declare mock as a dependency
+# via tests_require so it is not available on all supervisor installs.  the
+# modules imported in supervisor.compat must always be available.
+
+try: # pragma: no cover
+    from unittest.mock import Mock, patch, sentinel
+except ImportError: # pragma: no cover
+    from mock import Mock, patch, sentinel
+
+try: # pragma: no cover
+    import unittest.mock as mock
+except ImportError: # pragma: no cover
+    import mock
+
+class DummyOptions:
     loglevel = 20
+    minfds = 5
+
+    chdir_exception = None
+    fork_exception = None
+    execv_exception = None
+    kill_exception = None
+    make_pipes_exception = None
+    remove_exception = None
+    write_exception = None
 
     def __init__(self):
         self.identifier = 'supervisor'
@@ -53,11 +75,8 @@ class DummyOptions:
         self.setuid_msg = None
         self.privsdropped = None
         self.logs_reopened = False
-        self.environment_processed = False
         self.write_accept = None
-        self.write_error = None
         self.tempfile_name = '/foo/bar'
-        self.remove_error = None
         self.removed = []
         self.existing = []
         self.openreturn = None
@@ -67,9 +86,9 @@ class DummyOptions:
         self.parse_infos = []
         self.serverurl = 'http://localhost:9001'
         self.changed_directory = False
-        self.chdir_error = None
         self.umaskset = None
         self.poller = DummyPoller(self)
+        self.silent = False
 
     def getLogger(self, *args, **kw):
         logger = DummyLogger()
@@ -129,8 +148,8 @@ class DummyOptions:
         return self.waitpid_return
 
     def kill(self, pid, sig):
-        if self.kill_error:
-            raise OSError(self.kill_error)
+        if self.kill_exception is not None:
+            raise self.kill_exception
         self.kills[pid] = sig
 
     def stat(self, filename):
@@ -150,11 +169,9 @@ class DummyOptions:
             raise NotFound('bad filename')
 
     def make_pipes(self, stderr=True):
-        if self.make_pipes_error:
-            raise OSError(self.make_pipes_error)
-        pipes = {}
-        pipes['child_stdin'], pipes['stdin'] = (3, 4)
-        pipes['stdout'], pipes['child_stdout'] = (5, 6)
+        if self.make_pipes_exception is not None:
+            raise self.make_pipes_exception
+        pipes = {'child_stdin': 3, 'stdin': 4, 'stdout': 5, 'child_stdout': 6}
         if stderr:
             pipes['stderr'], pipes['child_stderr'] = (7, 8)
         else:
@@ -162,8 +179,8 @@ class DummyOptions:
         return pipes
 
     def write(self, fd, chars):
-        if self.write_error:
-            raise OSError(self.write_error)
+        if self.write_exception is not None:
+            raise self.write_exception
         if self.write_accept:
             chars = chars[self.write_accept]
         data = self.written.setdefault(fd, '')
@@ -172,8 +189,8 @@ class DummyOptions:
         return len(chars)
 
     def fork(self):
-        if self.fork_error:
-            raise OSError(self.fork_error)
+        if self.fork_exception is not None:
+            raise self.fork_exception
         return self.forkpid
 
     def close_fd(self, fd):
@@ -196,11 +213,8 @@ class DummyOptions:
 
     def execve(self, filename, argv, environment):
         self.execve_called = True
-        if self.execv_error:
-            if self.execv_error == 1:
-                raise OSError(self.execv_error)
-            else:
-                raise RuntimeError(self.execv_error)
+        if self.execv_exception is not None:
+            raise self.execv_exception
         self.execv_args = (filename, argv)
         self.execv_environment = environment
 
@@ -215,16 +229,12 @@ class DummyOptions:
     def reopenlogs(self):
         self.logs_reopened = True
 
-    def process_environment(self):
-        self.environment_processed = True
-
     def mktempfile(self, prefix, suffix, dir):
         return self.tempfile_name
 
     def remove(self, path):
-        import os
-        if self.remove_error:
-            raise os.error(self.remove_error)
+        if self.remove_exception is not None:
+            raise self.remove_exception
         self.removed.append(path)
 
     def exists(self, path):
@@ -238,14 +248,16 @@ class DummyOptions:
         return open(name, mode)
 
     def chdir(self, dir):
-        if self.chdir_error:
-            raise OSError(self.chdir_error)
+        if self.chdir_exception is not None:
+            raise self.chdir_exception
         self.changed_directory = True
 
     def setumask(self, mask):
         self.umaskset = mask
 
 class DummyLogger:
+    level = None
+
     def __init__(self):
         self.reopened = False
         self.removed = False
@@ -263,6 +275,9 @@ class DummyLogger:
             msg = msg % kw
         self.data.append(msg)
 
+    def addHandler(self, handler):
+        handler.close()
+
     def reopen(self):
         self.reopened = True
     def close(self):
@@ -321,8 +336,9 @@ class DummySocket:
         return 'dummy socket'
 
 class DummySocketConfig:
-    def __init__(self, fd):
+    def __init__(self, fd, backlog=128):
         self.fd = fd
+        self.backlog = backlog
         self.url = 'unix:///sock'
 
     def addr(self):
@@ -334,6 +350,9 @@ class DummySocketConfig:
     def __ne__(self, other):
         return not self.__eq__(other)
 
+    def get_backlog(self):
+        return self.backlog
+
     def create_and_bind(self):
         return DummySocket(self.fd)
 
@@ -347,7 +366,10 @@ class DummySocketManager:
     def get_socket(self):
         return DummySocket(self._config.fd)
 
-class DummyProcess:
+@functools.total_ordering
+class DummyProcess(object):
+    write_exception = None
+
     # Initial state; overridden by instance variables
     pid = 0 # Subprocess pid; 0 when not running
     laststart = 0 # Last time the subprocess was started; 0 if never
@@ -386,11 +408,11 @@ class DummyProcess:
         self.error_at_clear = False
         self.killed_with = None
         self.drained = False
-        self.stdout_buffer = ''
-        self.stderr_buffer = ''
-        self.stdout_logged = ''
-        self.stderr_logged = ''
-        self.stdin_buffer = ''
+        self.stdout_buffer = as_bytes('')
+        self.stderr_buffer = as_bytes('')
+        self.stdout_logged = as_bytes('')
+        self.stderr_logged = as_bytes('')
+        self.stdin_buffer = as_bytes('')
         self.pipes = {}
         self.rpipes = {}
         self.dispatchers = {}
@@ -400,7 +422,6 @@ class DummyProcess:
         self.input_fd_drained = None
         self.output_fd_drained = None
         self.transitioned = False
-        self.write_error = None
 
     def reopenlogs(self):
         self.logs_reopened = True
@@ -437,9 +458,6 @@ class DummyProcess:
     def drain(self):
         self.drained = True
 
-    def __cmp__(self, other):
-        return cmp(self.config.priority, other.config.priority)
-
     def readable_fds(self):
         return []
 
@@ -472,13 +490,19 @@ class DummyProcess:
         self.input_fd_drained = fd
 
     def write(self, chars):
-        if self.write_error:
-            raise OSError(self.write_error)
+        if self.write_exception is not None:
+            raise self.write_exception
         self.stdin_buffer += chars
 
     def transition(self):
         self.transitioned = True
 
+    def __eq__(self, other):
+        return self.config.priority == other.config.priority
+
+    def __lt__(self, other):
+        return self.config.priority < other.config.priority
+
 class DummyPConfig:
     def __init__(self, options, name, command, directory=None, umask=None,
                  priority=999, autostart=True,
@@ -486,12 +510,14 @@ class DummyPConfig:
                  uid=None, stdout_logfile=None, stdout_capture_maxbytes=0,
                  stdout_events_enabled=False,
                  stdout_logfile_backups=0, stdout_logfile_maxbytes=0,
+                 stdout_syslog=False,
                  stderr_logfile=None, stderr_capture_maxbytes=0,
                  stderr_events_enabled=False,
                  stderr_logfile_backups=0, stderr_logfile_maxbytes=0,
+                 stderr_syslog=False,
                  redirect_stderr=False,
                  stopsignal=None, stopwaitsecs=10, stopasgroup=False, killasgroup=False,
-                 exitcodes=(0,2), environment=None, serverurl=None):
+                 exitcodes=(0,), environment=None, serverurl=None):
         self.options = options
         self.name = name
         self.command = command
@@ -506,11 +532,13 @@ class DummyPConfig:
         self.stdout_events_enabled = stdout_events_enabled
         self.stdout_logfile_backups = stdout_logfile_backups
         self.stdout_logfile_maxbytes = stdout_logfile_maxbytes
+        self.stdout_syslog = stdout_syslog
         self.stderr_logfile = stderr_logfile
         self.stderr_capture_maxbytes = stderr_capture_maxbytes
         self.stderr_events_enabled = stderr_events_enabled
         self.stderr_logfile_backups = stderr_logfile_backups
         self.stderr_logfile_maxbytes = stderr_logfile_maxbytes
+        self.stderr_syslog = stderr_syslog
         self.redirect_stderr = redirect_stderr
         if stopsignal is None:
             import signal
@@ -526,6 +554,9 @@ class DummyPConfig:
         self.autochildlogs_created = False
         self.serverurl = serverurl
 
+    def get_path(self):
+        return ["/bin", "/usr/bin", "/usr/local/bin"]
+
     def create_autochildlogs(self):
         self.autochildlogs_created = True
 
@@ -562,11 +593,10 @@ def makeExecutable(file, substitutions=None):
     for key in substitutions.keys():
         data = data.replace('<<%s>>' % key.upper(), substitutions[key])
 
-    tmpnam = tempfile.mktemp(prefix=last)
-    f = open(tmpnam, 'w')
-    f.write(data)
-    f.close()
-    os.chmod(tmpnam, int('755', 8))
+    with tempfile.NamedTemporaryFile(prefix=last, delete=False) as f:
+        tmpnam = f.name
+        f.write(data)
+    os.chmod(tmpnam, 0o755)
     return tmpnam
 
 def makeSpew(unkillable=False):
@@ -636,7 +666,10 @@ class DummyRequest(object):
         del self.headers[header]
 
     def has_key(self, header):
-        return self.headers.has_key(header)
+        return header in self.headers
+
+    def __contains__(self, item):
+        return item in self.headers
 
     def done(self):
         self._done = True
@@ -722,41 +755,42 @@ class DummySupervisorRPCNamespace:
     def getPID(self):
         return 42
 
-    def readProcessStdoutLog(self, name, offset, length):
+    def _read_log(self, channel, name, offset, length):
         from supervisor import xmlrpc
-        import xmlrpclib
         if name == 'BAD_NAME':
-            raise xmlrpclib.Fault(xmlrpc.Faults.BAD_NAME, 'BAD_NAME')
+            raise Fault(xmlrpc.Faults.BAD_NAME, 'BAD_NAME')
         elif name == 'FAILED':
-            raise xmlrpclib.Fault(xmlrpc.Faults.FAILED, 'FAILED')
+            raise Fault(xmlrpc.Faults.FAILED, 'FAILED')
         elif name == 'NO_FILE':
-            raise xmlrpclib.Fault(xmlrpc.Faults.NO_FILE, 'NO_FILE')
-        a = 'output line\n' * 10
+            raise Fault(xmlrpc.Faults.NO_FILE, 'NO_FILE')
+        a = (channel + ' line\n') * 10
         return a[offset:]
 
+    def readProcessStdoutLog(self, name, offset, length):
+        return self._read_log('stdout', name, offset, length)
     readProcessLog = readProcessStdoutLog
-    readProcessStderrLog = readProcessStdoutLog
+
+    def readProcessStderrLog(self, name, offset, length):
+        return self._read_log('stderr', name, offset, length)
 
     def getAllProcessInfo(self):
         return self.all_process_info
 
     def getProcessInfo(self, name):
         from supervisor import xmlrpc
-        import xmlrpclib
         for i in self.all_process_info:
             if i['name']==name:
                 info=i
                 return info
         if name == 'BAD_NAME':
-            raise xmlrpclib.Fault(xmlrpc.Faults.BAD_NAME, 'BAD_NAME')
+            raise Fault(xmlrpc.Faults.BAD_NAME, 'BAD_NAME')
         if name == 'FAILED':
-            raise xmlrpclib.Fault(xmlrpc.Faults.FAILED, 'FAILED')
+            raise Fault(xmlrpc.Faults.FAILED, 'FAILED')
         if name == 'NO_FILE':
-            raise xmlrpclib.Fault(xmlrpc.Faults.NO_FILE, 'NO_FILE')
+            raise Fault(xmlrpc.Faults.NO_FILE, 'NO_FILE')
 
     def startProcess(self, name):
         from supervisor import xmlrpc
-        from xmlrpclib import Fault
         if name == 'BAD_NAME:BAD_NAME':
             raise Fault(xmlrpc.Faults.BAD_NAME, 'BAD_NAME:BAD_NAME')
         if name == 'BAD_NAME':
@@ -776,7 +810,7 @@ class DummySupervisorRPCNamespace:
 
     def startProcessGroup(self, name):
         from supervisor import xmlrpc
-        from xmlrpclib import Fault
+        from supervisor.compat import Fault
         if name == 'BAD_NAME':
             raise Fault(xmlrpc.Faults.BAD_NAME, 'BAD_NAME')
         return [
@@ -804,7 +838,7 @@ class DummySupervisorRPCNamespace:
 
     def stopProcessGroup(self, name):
         from supervisor import xmlrpc
-        from xmlrpclib import Fault
+        from supervisor.compat import Fault
         if name == 'BAD_NAME':
             raise Fault(xmlrpc.Faults.BAD_NAME, 'BAD_NAME')
         return [
@@ -818,7 +852,6 @@ class DummySupervisorRPCNamespace:
 
     def stopProcess(self, name):
         from supervisor import xmlrpc
-        from xmlrpclib import Fault
         if name == 'BAD_NAME:BAD_NAME':
             raise Fault(xmlrpc.Faults.BAD_NAME, 'BAD_NAME:BAD_NAME')
         if name == 'BAD_NAME':
@@ -847,7 +880,6 @@ class DummySupervisorRPCNamespace:
         if self._restartable:
             self._restarted = True
             return
-        from xmlrpclib import Fault
         from supervisor import xmlrpc
         raise Fault(xmlrpc.Faults.SHUTDOWN_STATE, '')
 
@@ -855,7 +887,6 @@ class DummySupervisorRPCNamespace:
         if self._restartable:
             self._shutdown = True
             return
-        from xmlrpclib import Fault
         from supervisor import xmlrpc
         raise Fault(xmlrpc.Faults.SHUTDOWN_STATE, '')
 
@@ -863,7 +894,6 @@ class DummySupervisorRPCNamespace:
         return [[['added'], ['changed'], ['removed']]]
 
     def addProcessGroup(self, name):
-        from xmlrpclib import Fault
         from supervisor import xmlrpc
         if name == 'ALREADY_ADDED':
             raise Fault(xmlrpc.Faults.ALREADY_ADDED, '')
@@ -879,7 +909,6 @@ class DummySupervisorRPCNamespace:
             self.processes = [name]
 
     def removeProcessGroup(self, name):
-        from xmlrpclib import Fault
         from supervisor import xmlrpc
         if name == 'STILL_RUNNING':
             raise Fault(xmlrpc.Faults.STILL_RUNNING, '')
@@ -890,7 +919,6 @@ class DummySupervisorRPCNamespace:
         self.processes.remove(name)
 
     def clearProcessStdoutLog(self, name):
-        from xmlrpclib import Fault
         from supervisor import xmlrpc
         if name == 'BAD_NAME':
             raise Fault(xmlrpc.Faults.BAD_NAME, 'BAD_NAME')
@@ -928,13 +956,11 @@ class DummySupervisorRPCNamespace:
 
     def readLog(self, whence, offset):
         if self._readlog_error:
-            from xmlrpclib import Fault
             raise Fault(self._readlog_error, '')
         return 'mainlogdata'
 
     def signalProcessGroup(self, name, signal):
         from supervisor import xmlrpc
-        from xmlrpclib import Fault
         if name == 'BAD_NAME':
             raise Fault(xmlrpc.Faults.BAD_NAME, 'BAD_NAME')
         return [
@@ -950,7 +976,6 @@ class DummySupervisorRPCNamespace:
 
     def signalProcess(self, name, signal):
         from supervisor import xmlrpc
-        from xmlrpclib import Fault
         if signal == 'BAD_SIGNAL':
             raise Fault(xmlrpc.Faults.BAD_SIGNAL, 'BAD_SIGNAL')
         if name == 'BAD_NAME:BAD_NAME':
@@ -1008,7 +1033,8 @@ class DummyFCGIGroupConfig(DummyPGroupConfig):
         DummyPGroupConfig.__init__(self, options, name, priority, pconfigs)
         self.socket_config = socket_config
 
-class DummyProcessGroup:
+@functools.total_ordering
+class DummyProcessGroup(object):
     def __init__(self, config):
         self.config = config
         self.transitioned = False
@@ -1032,6 +1058,15 @@ class DummyProcessGroup:
     def get_dispatchers(self):
         return self.dispatchers
 
+    def __lt__(self, other):
+        return self.config.priority < other.config.priority
+
+    def __eq__(self, other):
+        return self.config.priority == other.config.priority
+
+    def reopenlogs(self):
+        self.logs_reopened = True
+
 class DummyFCGIProcessGroup(DummyProcessGroup):
 
     def __init__(self, config):
@@ -1062,14 +1097,16 @@ class PopulatedDummySupervisor(DummySupervisor):
         self.reaped = True
 
 class DummyDispatcher:
+    flush_exception = None
+
     write_event_handled = False
     read_event_handled = False
     error_handled = False
     logs_reopened = False
     logs_removed = False
     closed = False
-    flush_error = None
     flushed = False
+
     def __init__(self, readable=False, writable=False, error=False):
         self._readable = readable
         self._writable = writable
@@ -1101,37 +1138,44 @@ class DummyDispatcher:
     def close(self):
         self.closed = True
     def flush(self):
-        if self.flush_error:
-            raise OSError(self.flush_error)
+        if self.flush_exception:
+            raise self.flush_exception
         self.flushed = True
 
 class DummyStream:
-    def __init__(self, error=None):
+    def __init__(self, error=None, fileno=20):
         self.error = error
         self.closed = False
         self.flushed = False
-        self.written = ''
+        self.written = b''
+        self._fileno = fileno
     def close(self):
         if self.error:
             raise self.error
         self.closed = True
     def flush(self):
+        if self.error:
+            raise self.error
         self.flushed = True
     def write(self, msg):
         if self.error:
-            raise self.error
-        self.written +=msg
+            error = self.error
+            self.error = None
+            raise error
+        self.written += as_bytes(msg)
     def seek(self, num, whence=0):
         pass
     def tell(self):
         return len(self.written)
+    def fileno(self):
+        return self._fileno
 
 class DummyEvent:
     def __init__(self, serial='abc'):
         if serial is not None:
             self.serial = serial
 
-    def __str__(self):
+    def payload(self):
         return 'dummy event'
 
 class DummyPoller:
